/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

typedef enum	ptype_s		ptype_t;
typedef struct	particle_s	particle_t;

enum ptype_s
{
	pt_fire,
	pt_explode
};

struct particle_s
{
	vec3_t				velocity;
	vec3_t				origin;
	
	float				die;
	float				scale;
	float				alpha;

	byte				colorred;
	byte				colorgreen;
	byte				colorblue;

	ptype_t				type;

	particle_t			*next;
	particle_t			*prev;
};

int		part_tex;

particle_t	*active_particles, *free_particles, *particles;

extern byte	particle[32][32];

void R_InitParticleTexture (void)
{
	int		x, y;
	byte	data[32][32][4];

	for (x=0 ; x<32 ; x++)
	{
		for (y=0 ; y<32 ; y++)
		{
			data[x][y][0]	= 255;
			data[x][y][1]	= 255;
			data[x][y][2]	= 255;
			data[x][y][3]	= particle[x][y];
		}
	}
	part_tex = GL_LoadTexture ("particle", 32, 32, &data[0][0][0], true, true, 4);
}

/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int	i;

	particles = (particle_t *) Hunk_AllocName (4096 * sizeof(particle_t), "particles");

	memset (particles, 0, 4096 * sizeof(particle_t));

	for (i=0; i<4096; ++i)
	{
		particles[i].prev	= &particles[i-1];
		particles[i].next	= &particles[i+1];
	}

	particles[0].prev		= NULL;
	particles[4095].next	= NULL;

	free_particles			= &particles[0];
	active_particles		= NULL;
}

/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;

	free_particles			= &particles[0];
	active_particles		= NULL;

	memset (particles, 0, 4096 * sizeof(particle_t));

	for (i=0; i<4096; ++i)
	{
		particles[i].prev	= &particles[i-1];
		particles[i].next	= &particles[i+1];
	}

	particles[0].prev		= NULL;
	particles[4095].next	= NULL;
}

particle_t* addParticle()
{
	particle_t*	p	= free_particles;

	if (!p)
	{
		return NULL;
	}

	free_particles	= p->next;

	if (free_particles)
	{
		free_particles->prev	= NULL;
	}

	p->next			= active_particles;

	if (active_particles)
	{
		active_particles->prev	= p;
	}

	active_particles			= p;

	return p;
}

particle_t* remParticle(particle_t*	p)
{
	particle_t*	next;

	if (!p)
	{
		return NULL;
	}

	if (p == active_particles)
	{
		active_particles	= p->next;
	}

	if (p->next)
	{
		p->next->prev		= p->prev;
	}

	if (p->prev)
	{
		p->prev->next		= p->next;
	}

	if (free_particles)
	{
		free_particles->prev	= p;
	}

	next					= p->next;

	p->next					= free_particles;

	free_particles			= p;

	p->prev	= NULL;

	return next;
}

	vec3_t				velocity;
	vec3_t				origin;
	
	float				die;
	float				scale;
	float				alpha;

	byte				colorred;
	byte				colorgreen;
	byte				colorblue;

	ptype_t				type;

	particle_t			*next;
	particle_t			*prev;
	
/*
========================
R_ParticleExplosion_Ship
========================
*/
void R_ParticleExplosion_Ship (vec3_t origin)
{
	int			i, contents;
	particle_t	*p;
	byte		*color;

	contents = Mod_PointInLeaf(origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	for (i=0 ; i<128 ; i++)
	{
		p = addParticle();
		if(!p)
		{
			return;
		}

		color				= (byte *) &d_8to24table[(int)(rand() & 7) + 232];

		p->velocity[0]		= (rand() & 31) - 16;
		p->velocity[1]		= (rand() & 31) - 16;
		p->velocity[2]		= (rand() & 31) - 16;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= 3;
		p->alpha			= 200;

		p->colorred			= color[0];
		p->colorgreen		= color[1];
		p->colorblue		= color[2];

		p->type				= pt_fire;
	}

	for (i=0 ; i<256 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;

		p->colorred			= 255;
		p->colorgreen		= 243;
		p->colorblue		= 27;

		p->type				= pt_explode;
	}
}

/*
=======================
R_ParticleExplosion_Ion
=======================
*/
void R_ParticleExplosion_Ion (vec3_t origin)
{
	int			i, contents;
	particle_t	*p;
	byte		*color;

	contents = Mod_PointInLeaf(origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	for (i=0 ; i<128 ; i++)
	{
		p = addParticle();
		if(!p)
		{
			return;
		}

		color				= (byte *) &d_8to24table[(int)(rand() & 7) + 232];

		p->velocity[0]		= (rand() & 31) - 16;
		p->velocity[1]		= (rand() & 31) - 16;
		p->velocity[2]		= (rand() & 31) - 16;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= 3;
		p->alpha			= 200;

		p->colorred			= color[0];
		p->colorgreen		= color[1];
		p->colorblue		= color[2];

		p->type				= pt_fire;
	}

	for (i=0 ; i<256 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;

		p->colorred			= 255;
		p->colorgreen		= 243;
		p->colorblue		= 27;

		p->type				= pt_explode;
	}
}

/*
===========================
R_ParticleExplosion_Torpedo
===========================
*/
void R_ParticleExplosion_Torpedo (vec3_t origin, int colorStart, int colorLength)
{
	int			i,contents;
	particle_t	*p;
	byte		*color;

	contents = Mod_PointInLeaf(origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	color = (byte *) &d_8to24table[colorStart];

	for (i=0 ; i<128 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]		= (rand() & 31) - 16;
		p->velocity[1]		= (rand() & 31) - 16;
		p->velocity[2]		= (rand() & 31) - 16;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= 3;
		p->alpha			= 200;

		p->colorred			= color[0];
		p->colorgreen		= color[1];
		p->colorblue		= color[2];

		p->type				= pt_fire;
	}

	for (i=0 ; i<256 ; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]		= (rand() & 511) - 256;
		p->velocity[1]		= (rand() & 511) - 256;
		p->velocity[2]		= (rand() & 511) - 256;

		p->origin[0]		= origin[0];
		p->origin[1]		= origin[1];
		p->origin[2]		= origin[2];

		p->die				= cl.time + 5;
		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;

		p->colorred			= color[0];
		p->colorgreen		= color[1];
		p->colorblue		= color[2];

		p->type				= pt_explode;
	}
}

/*
=============
R_SparkShower
=============
*/
void R_SparkShower (vec3_t origin, vec3_t direction)
{
	int			i, contents;
	particle_t	*p;

	contents = Mod_PointInLeaf(origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	p	= addParticle();
	if (!free_particles)
	{
		return;
	}

	p->velocity[0]		= (rand() & 7) - 4;
	p->velocity[1]		= (rand() & 7) - 4;
	p->velocity[2]		= (rand() & 7) - 4;


	p->origin[0]		= origin[0];
	p->origin[1]		= origin[1];
	p->origin[2]		= origin[2];

	p->die				= cl.time + 5;
	p->scale			= 2;
	p->alpha			= 200;

	p->colorred			= 187;
	p->colorgreen		= 187;
	p->colorblue		= 187;

	p->type				= pt_fire;

	for (i=0 ; i<16; i++)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]		= (rand() & 127) - 64;
		p->velocity[1]		= (rand() & 127) - 64;
		p->velocity[2]		= (rand() & 127) - 64;

		p->origin[0]		= origin[0] + ((rand() & 7) - 4);
		p->origin[1]		= origin[1] + ((rand() & 7) - 4);
		p->origin[2]		= origin[2] + ((rand() & 7) - 4);

		p->die				= cl.time + 5;
		p->scale			= (rand() & 3) +1;
		p->alpha			= 200;

		p->colorred			= 255;
		p->colorgreen		= 243;
		p->colorblue		= 147;

		p->type				= pt_explode;
	}
}

void R_RocketTrail (entity_t *ent)
{
	int			contents;
	particle_t	*p;

	contents = Mod_PointInLeaf (ent->origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	if (cl.time > ent->time_left)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]	= (rand() & 7) - 4;
		p->velocity[1]	= (rand() & 7) - 4;
		p->velocity[2]	= (rand() & 7) - 4;

		p->origin[0]	= ent->origin[0];
		p->origin[1]	= ent->origin[1];
		p->origin[2]	= ent->origin[2];

		p->die			= cl.time + 5;
		p->scale		= 2;
		p->alpha		= 200;

		p->colorred		= 187;
		p->colorgreen	= 187;
		p->colorblue	= 187;

		p->type			= pt_fire;

		ent->time_left	= cl.time + 0.05;
	}
}

void R_DebrisTrail (entity_t *ent)
{
	int			contents;
	particle_t	*p;

	if (ent->debris_smoke <= 0)
		return;

	contents = Mod_PointInLeaf (ent->origin, cl.worldmodel)->contents;

	if (contents != CONTENTS_EMPTY)
		return;

	if (cl.time > ent->time_left)
	{
		p	= addParticle();
		if(!p)
		{
			return;
		}

		p->velocity[0]	= (rand() & 7) - 4;
		p->velocity[1]	= (rand() & 7) - 4;
		p->velocity[2]	= (rand() & 7) - 4;

		p->origin[0]	= ent->origin[0];
		p->origin[1]	= ent->origin[1];
		p->origin[2]	= ent->origin[2];

		p->die			= cl.time + 5;
		p->scale		= 2;
		p->alpha		= ent->debris_smoke * 28;

		p->colorred		= 187;
		p->colorgreen	= 187;
		p->colorblue	= 187;	
		
		p->type			= pt_fire;

		ent->time_left		= cl.time + 0.05;
		ent->debris_smoke	-= 0.1f;
	}
}

/*
===============
R_DrawParticles
===============
*/

extern	cvar_t	sv_gravity;

void R_DrawParticles (void)
{
	particle_t	*p;

	float		frametime		= cl.time - cl.oldtime;

	float		ft12			= frametime * 12;
	float		ft128			= frametime * 128;

	vec3_t		up				= {vup[0]    * 0.75f, vup[1]    * 0.75f, vup[2]    * 0.75f};
	vec3_t		right			= {vright[0] * 0.75f, vright[1] * 0.75f, vright[2] * 0.75f};

	vec3_t		coord[4];

	if (!active_particles)
		return;

	VectorAdd      (up, right, coord[0]);
	VectorSubtract (right, up, coord[1]);
	VectorNegate   (coord[0], coord[2]);
	VectorNegate   (coord[1], coord[3]);

	glTexEnvf	  (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDepthMask	  (false);
	glBlendFunc	  (GL_SRC_ALPHA, GL_ONE);
    glBindTexture (GL_TEXTURE_2D, part_tex);

	p = active_particles;

	while(p)
	{
		p->origin[0] += p->velocity[0] * frametime;
		p->origin[1] += p->velocity[1] * frametime;
		p->origin[2] += p->velocity[2] * frametime;

		if (p->type == pt_fire)
		{
			p->scale += ft12;
		}

		p->alpha -= ft128;

		if (p->alpha < 0)
		{
			p = remParticle(p);
			continue;
		}

		glColor4ub (p->colorred, p->colorgreen, p->colorblue, (byte)p->alpha);

		glPushMatrix ();
	
		glTranslatef (p->origin[0], p->origin[1], p->origin[2]);
		glScalef (p->scale, p->scale, p->scale);
			
		glBegin (GL_QUADS);

			glTexCoord2f (0,1);		glVertex3fv (coord[0]);
			glTexCoord2f (0,0);		glVertex3fv (coord[1]);
			glTexCoord2f (1,0);		glVertex3fv (coord[2]);
			glTexCoord2f (1,1);		glVertex3fv (coord[3]);

		glEnd ();

		glPopMatrix ();

		p = p->next;
	}

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (true);
	glTexEnvf	(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glColor4f	(1,1,1,1);
}
