/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

entity_t	r_worldentity;

vec3_t		modelorg;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value

cvar_t	r_norefresh			= {"r_norefresh",		"0"};
cvar_t	r_speeds			= {"r_speeds",			"0"};
cvar_t	r_dynamic			= {"r_dynamic",			"1"};
cvar_t	r_novis				= {"r_novis",			"0"};

cvar_t	gl_finish			= {"gl_finish",			"0"};
cvar_t	gl_keeptjunctions	= {"gl_keeptjunctions",	"0"};

// Tomaz - Fog Begin
cvar_t  gl_fogenable		= {"gl_fogenable",		"0",		true}; 
cvar_t  gl_fogstart			= {"gl_fogstart",		"50.0",		true}; 
cvar_t  gl_fogend			= {"gl_fogend",			"800.0",	true}; 
cvar_t  gl_fogred			= {"gl_fogred",			"0.6",		true};
cvar_t  gl_foggreen			= {"gl_foggreen",		"0.5",		true}; 
cvar_t  gl_fogblue			= {"gl_fogblue",		"0.4",		true}; 
// Tomaz - Fog End

/*
=========
R_CullBox
=========
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	if ((BoxOnPlaneSide (mins, maxs, &frustum[0]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[1]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[2]) == 2) ||
		(BoxOnPlaneSide (mins, maxs, &frustum[3]) == 2))
		return true;
	return false;
}

void R_RotateForEntity (entity_t *e)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
	glRotatef (-e->angles[0],  0, 1, 0);
	glRotatef (e->angles[2],  1, 0, 0);
}

// Tomaz - Model Transform Interpolation Begin

/*
=============
R_BlendedRotateForEntity
=============
*/
void R_BlendedRotateForEntity (entity_t *e)
{
	float timepassed;
	float blend;
	vec3_t d;
	int i;

	// positional interpolation

	timepassed = realtime - e->translate_start_time; 

	if (e->translate_start_time == 0 || timepassed > 1)
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin, e->origin1);
		VectorCopy (e->origin, e->origin2);
	}

	if (!VectorCompare (e->origin, e->origin2))
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin2, e->origin1);
		VectorCopy (e->origin,  e->origin2);
		blend = 0;
	}
	else
	{
		blend =  (timepassed * 10) * slowmo.value; 	// Tomaz - Speed

		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->origin2, e->origin1, d);

	glTranslatef (e->origin1[0] + (blend * d[0]), e->origin1[1] + (blend * d[1]), e->origin1[2] + (blend * d[2]));

	// orientation interpolation (Euler angles, yuck!)

	timepassed = realtime - e->rotate_start_time; 

	if (e->rotate_start_time == 0 || timepassed > 1)
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles, e->angles1);
		VectorCopy (e->angles, e->angles2);
	}

	if (!VectorCompare (e->angles, e->angles2))
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles2, e->angles1);
		VectorCopy (e->angles,  e->angles2);
		blend = 0;
	}
	else
	{
		blend = (timepassed * 10) * slowmo.value;	// Tomaz - Speed
 
		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->angles2, e->angles1, d);

	// always interpolate along the shortest path
	for (i = 0; i < 3; i++) 
	{
		if (d[i] > 180)
		{
			d[i] -= 360;
		}
		else if (d[i] < -180)
		{
			d[i] += 360;
		}
	}

	glRotatef ( e->angles1[1] + ( blend * d[1]),  0, 0, 1);
	glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
	glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
}
// Tomaz - Model Transform Interpolation End

/*
=============================================================

  ALIAS MODELS

=============================================================
*/

// Tomaz - Lit Support

// precalculated dot products for quantized angles
float	r_avertexnormal_dots[16][256] =
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

int		lastposenum;
int		lastposenum0;	// Tomaz - Model Animation Interpolation

extern	vec3_t lightcolor; // Tomaz - Lit Support

// Tomaz - Model Animation Interpolation Begin
/*
=============
GL_DrawAliasBlendedFrame
=============
*/
void GL_DrawAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e)
{
	float		l, blend;
	trivertx_t	*verts1, *verts2;
	int			count, pose, numposes, *order;
	vec3_t		d;

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("GL_DrawAliasBlendedFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose		= paliashdr->frames[frame].firstpose;
	numposes	= paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		e->frame_interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / e->frame_interval) % numposes;
	}
	else 
	{
		e->frame_interval = 0.1f;
	}

	if (e->pose2 != pose)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else
	{
		blend = ((realtime - e->frame_start_time)*slowmo.value) / e->frame_interval;

		if (cl.paused || blend > 1) 
			blend = 1;
	}

	lastposenum0 = e->pose1;
	lastposenum  = e->pose2;

	verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2  = verts1;

	verts1 += e->pose1 * paliashdr->poseverts;
	verts2 += e->pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;
        
		if (!count) 
			break;
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
		{
			glBegin (GL_TRIANGLE_STRIP);
		}
		do
		{
			l = shadedots[verts1->lightnormalindex];
			glColor3f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2]);

			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			VectorSubtract(verts2->v, verts1->v, d);

			// blend the vertex positions from each frame together
			glVertex3f (verts1->v[0] + (blend * d[0]), verts1->v[1] + (blend * d[1]), verts1->v[2] + (blend * d[2]));

			verts1++;
			verts2++;
		} 
		while 
			(--count);
		glEnd ();
	}
	glColor4f (1,1,1,1);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}
// Tomaz - Model Animation Interpolation End

extern	vec3_t			lightspot;

int R_LightPoint (vec3_t p);

/*
================
R_DrawAliasModel
================
*/
void R_DrawAliasModel (entity_t *e, qboolean cull, qboolean interpolate)
{
	int			lnum, anim;
	float		add;
	vec3_t		dist, mins, maxs;
	model_t		*clmodel;
	aliashdr_t	*paliashdr;

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	if (cull && R_CullBox (mins, maxs))
		return;

	VectorSubtract (r_origin, currententity->origin, modelorg);

	//
	// get lighting information
	//

	R_LightPoint(currententity->origin); // Tomaz - Lit Support

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (currententity->origin, cl_dlights[lnum].origin,	dist);
			add = cl_dlights[lnum].radius - Length(dist);

			// Tomaz - Lit Support Begin
			if (add > 0)
			{
				lightcolor[0] += add * cl_dlights[lnum].color[0];
				lightcolor[1] += add * cl_dlights[lnum].color[1];
				lightcolor[2] += add * cl_dlights[lnum].color[2];
			}
			// Tomaz - Lit Support End
		}
	}

	if (lightcolor[0] < 64)
		lightcolor[0] = 64;
	if (lightcolor[1] < 64)
		lightcolor[1] = 64;
	if (lightcolor[2] < 64)
		lightcolor[2] = 64;

	shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (0.0444444444))) & (15)];	// Tomaz - Speed

	// Tomaz - Lit Support Begin
	lightcolor[0] = lightcolor[0] * 0.005;
	lightcolor[1] = lightcolor[1] * 0.005;
	lightcolor[2] = lightcolor[2] * 0.005;
	// Tomaz - Lit Support End

	//
	// locate the proper data
	//

	paliashdr = (aliashdr_t *)Mod_Extradata (currententity->model);
	c_alias_polys += paliashdr->numtris;

	//
	// draw all the triangles
	//

	glPushMatrix ();

	if (interpolate)
		R_BlendedRotateForEntity (e);
	else
		R_RotateForEntity(e);

	glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
	glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);

	anim = (int)(cl.time*10) & 3;
    glBindTexture (GL_TEXTURE_2D, paliashdr->gl_texturenum[currententity->skinnum][anim]);

	GL_DrawAliasBlendedFrame (currententity->frame, paliashdr, currententity);

	glPopMatrix ();
}
//==================================================================================

void R_SetupBrushPolys (entity_t *e);
void R_DrawSpriteModel (entity_t *e);

/*
===============
R_SetupEntities
===============
*/
void R_SetupEntities (void)
{
	int		i;

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];
	
		switch (currententity->model->type)
		{
		case mod_alias:
		{
			if (currententity == &cl_entities[cl.viewentity])
			{
				currententity->angles[0] *= 0.3f;
			}

			R_DrawAliasModel  (currententity, true, true);
			break;
		}
		case mod_brush:
			R_SetupBrushPolys (currententity);
			break;
		case mod_sprite:
			R_DrawSpriteModel (currententity);
			break;
		default:
			break;
		}
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	if (cl.items & IT_INVISIBILITY)
		return;

	if (cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;
	if (!currententity->model)
		return;

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (0, 0.3);

	R_DrawAliasModel (currententity, false, false);

	glDepthRange (0, 1);
}

/*
===========
R_PolyBlend
===========
*/
void R_PolyBlend (void)
{
	if (!v_blend[3])
		return;

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef ( 90,  0, 0, 1);	    // put Z going up

	glColor4fv(v_blend);

	glBegin (GL_QUADS);
		glVertex3f (10,  100,  100);
		glVertex3f (10, -100,  100);
		glVertex3f (10, -100, -100);
		glVertex3f (10,  100, -100);
	glEnd ();

	glEnable (GL_TEXTURE_2D);
	glColor4f (1,1,1,1);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}

void R_SetFrustum (void)
{
	int		i;

	VectorAdd (vpn, vright, frustum[0].normal);
	VectorSubtract (vpn, vright, frustum[1].normal);

	VectorAdd (vpn, vup, frustum[2].normal);
	VectorSubtract (vpn, vup, frustum[3].normal);

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}

void R_AnimateLight (void);
void V_CalcBlend (void);

/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
	R_AnimateLight ();

	r_framecount++;

// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.origin, r_origin);

	AngleVectors (r_refdef.angles, vpn, vright, vup);

// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_CalcBlend ();

	c_brush_polys = 0;
	c_alias_polys = 0;
}

/*
=========
R_SetupGL
=========
*/
void R_SetupGL (void)
{
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

	glViewport (0, 0, vid.width, vid.height);
 
	glFrustum (-4, 4, -3, 3, 4, 6144);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);
    glRotatef ( 90,  0, 0, 1);
    glRotatef (-r_refdef.angles[2],  1, 0, 0);
    glRotatef (-r_refdef.angles[0],  0, 1, 0);
    glRotatef (-r_refdef.angles[1],  0, 0, 1);

    glTranslatef (-r_refdef.origin[0],  -r_refdef.origin[1],  -r_refdef.origin[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	glEnable(GL_DEPTH_TEST);
}

/*
=======
R_Clear
=======
*/
void R_Clear (void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void R_DrawWorld (void);
void R_DrawWaterSurfaces (void);
void R_DrawParticles (void);
void R_MarkLeaves (void);
/*
================
R_RenderView
================
*/
void R_RenderView (void)
{
	vec3_t		colors;	// Tomaz - Fog

	double	time1, time2;

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Host_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_FloatTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	if (gl_finish.value)
		glFinish ();

	R_Clear ();

	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_SetupEntities ();

	R_DrawParticles ();

	R_DrawViewModel ();

	R_DrawWaterSurfaces ();

	R_PolyBlend ();

	if (r_speeds.value)
	{
//		glFinish ();
		time2 = Sys_FloatTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}

	// Tomaz - Fog Begin
	glDisable(GL_FOG);

	if (gl_fogenable.value)
	{
		glFogi(GL_FOG_MODE, GL_LINEAR);
			colors[0] = gl_fogred.value;
			colors[1] = gl_foggreen.value;
			colors[2] = gl_fogblue.value; 
		glFogfv(GL_FOG_COLOR, colors); 
		glFogf(GL_FOG_START, gl_fogstart.value); 
		glFogf(GL_FOG_END, gl_fogend.value); 
		glEnable(GL_FOG);
	}
	// Tomaz - Fog End
}