/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include	"quakedef.h"
#include	"winquake.h"
#include	"resource.h"
#include	<commctrl.h>

qboolean	scr_skipupdate;

static DEVMODE	devmode;
static int		windowed_mouse;
extern qboolean	mouseactive;
static qboolean	vid_initialized		= false;
static qboolean vid_wassuspended	= false;

mode_t	mode;
RECT	WindowRect;

HDC			baseDC;
HWND		window;
HGLRC		baseRC;
HINSTANCE	hInstance;

viddef_t	vid;

unsigned	d_8to24table[256];

qboolean	gl_mtexable	= false;

modestate_t	modestate = MS_UNINIT;

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void AppActivate (qboolean fActive, qboolean minimize);
void ClearAllStates (void);
void GL_Init (void);

cvar_t		_windowed_mouse = {"_windowed_mouse","1", true};

int			window_center_x, window_center_y;

//====================================

void CenterWindow (void)
{
    int     CenterX, CenterY;

	CenterX = (GetSystemMetrics(SM_CXSCREEN) - vid.width)  * 0.5;
	CenterY = (GetSystemMetrics(SM_CYSCREEN) - vid.height) * 0.5;
	SetWindowPos (window, NULL, CenterX, CenterY, 0, 0,	SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

void CenterMouse (void)
{
	window_center_x = GetSystemMetrics(SM_CXSCREEN) * 0.5;
	window_center_y = GetSystemMetrics(SM_CYSCREEN) * 0.5;
}

void VID_SetMode (void)
{
	int						temp, width, height, PixelFormat;
	MSG						Msg;
	DWORD					WindowStyle;
	WNDCLASS				wc;
	PIXELFORMATDESCRIPTOR	pfd;

	// so Con_Printfs don't mess us up by forcing vid and snd updates
	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;

	/* Register the frame class */
	hInstance			= GetModuleHandle(NULL);

	memset(&wc, 0, sizeof(wc));

	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc		= (WNDPROC)MainWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= 0;
	wc.hCursor			= LoadCursor (NULL,IDC_ARROW);
	wc.hbrBackground	= NULL;
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= "StarWars: DogFighter";

    if (!RegisterClass (&wc))
		Sys_Error ("Couldn't register window class");

	WindowRect.left		= 0;
	WindowRect.top		= 0;
	WindowRect.right	= mode.width;
	WindowRect.bottom	= mode.height;

	vid.width			= mode.width;
	vid.height			= mode.height;

	// Set either the fullscreen or windowed mode

	if (mode.fullscreen)
	{
		memset (&devmode, 0, sizeof(devmode));

		devmode.dmFields		= DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		devmode.dmPelsWidth		= mode.width;
		devmode.dmPelsHeight	= mode.height;
		devmode.dmBitsPerPel	= mode.bpp;
		devmode.dmSize			= sizeof (devmode);

		if (ChangeDisplaySettings (&devmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
			Sys_Error ("Couldn't set fullscreen mode");

		modestate		= MS_FULL;
		WindowStyle		= WS_POPUP;
	}
	else
	{
		modestate		= MS_WINDOWED;
		WindowStyle		= WS_SYSMENU | WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_MINIMIZEBOX;
	}

	AdjustWindowRect (&WindowRect, WindowStyle, false);

	width	= WindowRect.right  - WindowRect.left;
	height	= WindowRect.bottom - WindowRect.top;

	// Create the window
	window = CreateWindow
		("StarWars: DogFighter",
		 "StarWars: DogFighter",
		 WindowStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
		 WindowRect.left,
		 WindowRect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 hInstance,
		 NULL);

	if (!window)
		Sys_Error ("Couldn't create window");

	CenterWindow ();
	ShowWindow   (window, SW_SHOWDEFAULT);
	UpdateWindow (window);

	SetForegroundWindow (window);

	while (PeekMessage (&Msg, NULL, 0, 0, PM_REMOVE))
	{
      	TranslateMessage (&Msg);
      	DispatchMessage  (&Msg);
	}

	Sleep (100);

	SetWindowPos (window, HWND_TOP, 0, 0, 0, 0, SWP_DRAWFRAME | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW | SWP_NOCOPYBITS);

	SetForegroundWindow (window);

	// Settings for the OpenGL window

	memset(&pfd, 0, sizeof(pfd));
	
	pfd.nSize			= sizeof(PIXELFORMATDESCRIPTOR);	// Size of this pixel format descriptor
	pfd.nVersion		= 1;								// The version of this data structure
	pfd.dwFlags			= PFD_DRAW_TO_WINDOW			    // Buffer supports drawing to window
						| PFD_SUPPORT_OPENGL				// Buffer supports OpenGL drawing
						| PFD_DOUBLEBUFFER;					// Supports double buffering
	pfd.iPixelType		= PFD_TYPE_RGBA;					// RGBA color format
	pfd.cColorBits		= mode.bpp;							// OpenGL color depth
	pfd.cDepthBits		= 32;								// Specifies the depth of the depth buffer
	pfd.iLayerType		= PFD_MAIN_PLANE;					// Main layer

	baseDC = GetDC (window);
	if (!baseDC)
		Sys_Error ("Couldn't create DC");

	PixelFormat = ChoosePixelFormat (baseDC, &pfd);
	if (!PixelFormat)
		Sys_Error ("Couldn't find pixelformat");

	if(!SetPixelFormat (baseDC, PixelFormat, &pfd))
		Sys_Error ("Couldn't set pixelformat");

	baseRC = wglCreateContext (baseDC);
	if (!baseRC)
		Sys_Error ("Couldn't create RC");

	if(!wglMakeCurrent (baseDC, baseRC))
		Sys_Error ("Couldn't activate DC and RC");

	scr_disabled_for_loading = temp;

	CenterMouse ();

	ShowWindow			(window,SW_SHOW);
	SetForegroundWindow	(window);
	SetFocus			(window);
	UpdateWindow		(window);

	ClearAllStates ();
}

//====================================

int		texture_mode = GL_LINEAR;

// Multitexture
#define		TEXTURE0_SGIS		0x835E
#define		TEXTURE1_SGIS		0x835F

#define		TEXTURE0_ARB		0x84C0
#define		TEXTURE1_ARB		0x84C1

GLenum		TEXTURE0;
GLenum		TEXTURE1;

void CheckMultiTextureExtensions(void) 
{
	const char *gl_extensions;

	gl_extensions	= glGetString (GL_EXTENSIONS);

	Con_Printf ("\nGL_VENDOR: %s\n", glGetString (GL_VENDOR));
	Con_Printf ("GL_RENDERER: %s\n\n", glGetString (GL_RENDERER));

	if (COM_CheckParm("-nomtex"))
	{
		Con_Printf("Multitexture Disabled\n");
		return;
	}

	if (strstr(gl_extensions, "GL_ARB_multitexture ")) 
	{
		Con_Printf("GL_ARB_multitexture enabled\n");
		MTexCoord2f		= (void *) wglGetProcAddress("glMultiTexCoord2fARB");
		SelectTexture	= (void *) wglGetProcAddress("glActiveTextureARB");
		gl_mtexable		= true;
		TEXTURE0		= TEXTURE0_ARB;
		TEXTURE1		= TEXTURE1_ARB;	
		return;
	} 
	
	if (strstr(gl_extensions, "GL_SGIS_multitexture ")) 
	{
		Con_Printf("GL_SGIS_multitexture enabled\n");
		MTexCoord2f		= (void *) wglGetProcAddress("glMTexCoord2fSGIS");
		SelectTexture	= (void *) wglGetProcAddress("glSelectTextureSGIS");
		gl_mtexable		= true;
		TEXTURE0		= TEXTURE0_SGIS;
		TEXTURE1		= TEXTURE1_SGIS;
		return;
	}
	Con_Printf("Multitexture Not Found\n");
}

/*
=======
GL_Init
=======
*/
void GL_Init (void)
{
	CheckMultiTextureExtensions ();

	glEnable		(GL_BLEND);
	glEnable		(GL_DITHER);
	glEnable		(GL_CULL_FACE);
	glEnable		(GL_TEXTURE_2D);
	glEnable		(GL_POINT_SMOOTH);
	glDisable		(GL_ALPHA_TEST);

	glHint			(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glCullFace		(GL_FRONT);
	glClearColor	(0,0,0,1);

	glAlphaFunc		(GL_GREATER, 0.5);
	glBlendFunc		(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthFunc		(GL_LEQUAL);	

	glTexEnvf		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glShadeModel	(GL_SMOOTH);
	glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);

	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

void GL_EndRendering (void)
{
	if (!scr_skipupdate || block_drawing)
		SwapBuffers(baseDC);

// handle the mouse state when windowed if that's changed
	if (modestate == MS_WINDOWED)
	{
		if (!_windowed_mouse.value) 
		{
			if (windowed_mouse)	
			{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
				windowed_mouse = false;
			}
		} 
		else 
		{
			windowed_mouse = true;
			if (key_dest == key_game && !mouseactive && ActiveApp) 
			{
				IN_ActivateMouse ();
				IN_HideMouse ();
			} 
			else if (mouseactive && key_dest != key_game) 
			{
				IN_DeactivateMouse ();
				IN_ShowMouse ();
			}
		}
	}
}

void VID_SetPalette (unsigned char *palette)
{
	byte			*pal;
	unsigned		r, g, b, v, *table;
	unsigned short	i;

//
// 8 8 8 encoding
//
	pal		= palette;
	table	= d_8to24table;

	for (i=0 ; i<256 ; i++)
	{
		r = pal[0];
		g = pal[1];
		b = pal[2];
		pal += 3;
		
		v = (255<<24) + (r<<0) + (g<<8) + (b<<16);
		*table++ = v;
	}
	d_8to24table[255] = 0x000000;	// 255 is transparent
}

BOOL	gammaworks;

void VID_SetDefaultMode (void)
{
	IN_DeactivateMouse ();
}

void VID_Shutdown (void)
{
	if (mode.fullscreen)
	{
		ChangeDisplaySettings (NULL, 0);
	}

	if (baseRC)
	{
		if (!wglMakeCurrent(NULL,NULL))
			Con_Printf ("Could Not Release RC And DC");

		if (!wglDeleteContext(baseRC))
			Con_Printf ("Could Not Release RC");
	}

	if (baseDC)
	{
		if (!ReleaseDC (window, baseDC))
			Con_Printf ("Could Not Release DC");
	}

	if (window)
	{
		ShowWindow (window, SW_HIDE);

		if (!DestroyWindow (window))
			Con_Printf ("Could Not Destroy window");
	}

	if (!UnregisterClass ("MiniRacer", hInstance))
		Con_Printf ("Could Not Unregister Class");

	if (mode.fullscreen)
	{
		ChangeDisplaySettings (NULL, 0);
		ShowCursor (true);
	}

	baseRC		= NULL;
	baseDC		= NULL;
	window		= NULL;
	hInstance	= NULL;
}

//==========================================================================

byte scantokey[128] = 
{ 
//  0		1		2		3		4		5		6			7
//  8		9		A		B		C		D		E			F
	0,		27,		'1',	'2',	'3',	'4',	'5',		'6',
	'7',	'8',	'9',	'0',	'-',	'=',	K_BACKSPACE,9,		// 0
	'q',	'w',    'e',    'r',	't',    'y',	'u',		'i',
	'o',	'p',	'[',	']',	13 ,    K_CTRL,	'a',		's',	// 1
	'd',	'f',    'g',	'h',	'j',    'k',    'l',		';',
	'\'',	'`',	K_SHIFT,'\\',	'z',    'x',    'c',		'v',	// 2
	'b',	'n',    'm',    ',',	'.',    '/',    K_SHIFT,	'*',
	K_ALT,	' ',	0  ,    K_F1,	K_F2,	K_F3,	K_F4,		K_F5,   // 3
	K_F6,	K_F7,	K_F8,	K_F9,	K_F10,	K_PAUSE,0,			K_HOME,
	K_UP,	K_PGUP,	'-',	K_LEFT,	'5',	K_RIGHT,'+',		K_END,	// 4
	K_DOWN,	K_PGDN,	K_INS,	K_DEL,	0,		0,		0,			K_F11,
	K_F12,	0,		0,		0,		0,		0,		0,			0,		// 5
	0,		0,		0,		0,		0  ,    0,		0,			0,
	0,		0,		0,		0,		0  ,    0,		0,			0,		// 6
	0,		0,		0,		0,		0  ,    0,		0,			0,
	0,		0,		0,		0,		0  ,    0,		0,			0		// 7
}; 

/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
	key = (key>>16)&255;
	if (key > 127)
		return 0;
	if (scantokey[key] == 0)
		Con_DPrintf("key 0x%02x has no translation\n", key);
	return scantokey[key];
}

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
	int		i;
	
// send an up event for each key, to make sure the server clears them all
	for (i=0 ; i<256 ; i++)
	{
		Key_Event (i, false);
	}

	Key_ClearStates ();
	IN_ClearStates ();
}

void AppActivate(BOOL fActive, BOOL minimize)
{
	static BOOL	sound_active;

	ActiveApp = fActive;
	Minimized = minimize;

// enable/disable sound on focus gain/loss
	if (!ActiveApp && sound_active)
	{
		S_BlockSound ();
		sound_active = false;
	}
	else if (ActiveApp && !sound_active)
	{
		S_UnblockSound ();
		sound_active = true;
	}

	if (fActive)
	{
		if (modestate == MS_FULL)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
			if (vid_wassuspended) 
			{
				vid_wassuspended = false;
				ChangeDisplaySettings (&devmode, CDS_FULLSCREEN);
				ShowWindow(window, SW_SHOWNORMAL);
				MoveWindow(window, 0, 0, devmode.dmPelsWidth, devmode.dmPelsHeight, false);	// Tomaz - Alt+Tab Fix
			}
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value && key_dest == key_game)
		{
			IN_ActivateMouse ();
			IN_HideMouse ();
		}
	}

	if (!fActive)
	{
		if (modestate == MS_FULL)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
			ChangeDisplaySettings (NULL, 0);
			vid_wassuspended = true;
		}
		else if ((modestate == MS_WINDOWED) && _windowed_mouse.value)
		{
			IN_DeactivateMouse ();
			IN_ShowMouse ();
		}
	}
}

/* main window procedure */
LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    LONG    lRet = 1;
	int		fActive, fMinimized, temp;
	extern unsigned int uiWheelMessage;

	if (uMsg == uiWheelMessage)
		uMsg = WM_MOUSEWHEEL;

    switch (uMsg)
    {
		case WM_SYSCHAR:
		case WM_SIZE:
		case WM_CREATE:
		case WM_MOVE:
		case WM_DESTROY:
			break;

		case WM_KILLFOCUS:
			if (modestate == MS_FULL)
				ShowWindow(window, SW_SHOWMINNOACTIVE);
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			Key_Event (MapKey(lParam), true);
			break;

		case WM_KEYUP:
		case WM_SYSKEYUP:
			Key_Event (MapKey(lParam), false);
			break;

	// this is complicated because Win32 seems to pack multiple mouse events into
	// one update sometimes, so we always check all states and look for events
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_MOUSEMOVE:
			temp = 0;

			if (wParam & MK_LBUTTON)
				temp |= 1;

			if (wParam & MK_RBUTTON)
				temp |= 2;

			if (wParam & MK_MBUTTON)
				temp |= 4;

			IN_MouseEvent (temp);

			break;

		// JACK: This is the mouse wheel with the Intellimouse
		// Its delta is either positive or neg, and we generate the proper
		// Event.
		case WM_MOUSEWHEEL: 
			if ((short) HIWORD(wParam) > 0) {
				Key_Event(K_MWHEELUP, true);
				Key_Event(K_MWHEELUP, false);
			} else {
				Key_Event(K_MWHEELDOWN, true);
				Key_Event(K_MWHEELDOWN, false);
			}
			break;

   	    case WM_CLOSE:
			if (MessageBox (window, "Are you sure you want to quit?", "Confirm Exit", MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
			{
				Sys_Quit ();
				PostQuitMessage(0);
			}

	        break;

		case WM_ACTIVATE:
			fActive = LOWORD(wParam);
			fMinimized = (BOOL) HIWORD(wParam);
			AppActivate (!(fActive == WA_INACTIVE), fMinimized);

			ClearAllStates ();

			break;

		case MM_MCINOTIFY:
//            lRet = CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
			break;

    	default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}

void Draw_InitTextures (void);
void Draw_Shutdown (void);

qboolean	changed_res = false;

void SwitchResolution (void)
{
	int		screenmode;

	mode.width		= 640;
	mode.height		= 480;
	mode.bpp		= 16;
	mode.fullscreen	= false;

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("  0  - 640  x 480 x 16 Window\n");
		Con_Printf ("  1  - 800  x 600 x 16 Window\n");
		Con_Printf ("  2  - 1024 x 768 x 16 Window\n");
		Con_Printf ("  3  - 640  x 480 x 32 Window\n");
		Con_Printf ("  4  - 800  x 600 x 32 Window\n");
		Con_Printf ("  5  - 1024 x 768 x 32 Window\n");
		Con_Printf ("  6  - 640  x 480 x 16 FullScreen\n");
		Con_Printf ("  7  - 800  x 600 x 16 FullScreen\n");
		Con_Printf ("  8  - 1024 x 768 x 16 FullScreen\n");
		Con_Printf ("  9  - 640  x 480 x 32 FullScreen\n");
		Con_Printf ("  10 - 800  x 600 x 32 FullScreen\n");
		Con_Printf ("  11 - 1024 x 768 x 32 FullScreen\n");
		return;
	}

	changed_res = true;

	screenmode = Q_atoi(Cmd_Argv(1));

	if (screenmode > 11)
		screenmode = 11;

	if (screenmode > 5)
	{
		mode.fullscreen	= true;
		screenmode -= 6;
	}

	if (screenmode > 2)
	{
		mode.bpp		= 32;
		screenmode -= 3;
	}

	if (screenmode == 0)
	{
		mode.width		= 640;
		mode.height		= 480;
	}

	if (screenmode == 1)
	{
		mode.width		= 800;
		mode.height		= 600;
	}

	if (screenmode == 2)
	{
		mode.width		= 1024;
		mode.height		= 768;
	}

	Draw_Shutdown ();

	VID_Shutdown ();

	VID_SetMode ();

	GL_Init ();

	Draw_InitTextures ();

	Cbuf_AddText ("restart\n");
}

void SaveResolution (void)
{
	FILE	*file;

	file = fopen("swd/screen.txt","w");

	if (file)
	{
		fprintf (file, "%i\n", mode.width);
		fprintf (file, "%i\n", mode.height);
		fprintf (file, "%i\n", mode.bpp);
		fprintf (file, "%i\n", mode.fullscreen);

		fprintf (file, "\n");
		fprintf (file, "// Width.\n");
		fprintf (file, "// Height.\n");
		fprintf (file, "// Bits Per Pixels.\n");
		fprintf (file, "// FullScreen\n");
		fprintf (file, "\n");
		fprintf (file, "// Resolution file made by SW:D.\n");
		fprintf (file, "// Do not edit.\n");
	}
}

/*
===================
VID_Init
===================
*/
void VID_Init (unsigned char *palette)
{
	FILE		*file;
	qboolean	fullscreen = true;

	mode.width		= 640;
	mode.height		= 480;
	mode.bpp		= 16;
	mode.fullscreen	= true;

	file = fopen("swd/screen.txt","rb");

	if (file)
	{
		fscanf (file, "%i", &mode.width);
		fscanf (file, "%i", &mode.height);
		fscanf (file, "%i", &mode.bpp);
		fscanf (file, "%i", &mode.fullscreen);

		fclose (file);
	}

	Cvar_RegisterVariable (&_windowed_mouse);
	Cmd_AddCommand ("resolution", SwitchResolution);

	vid.colormap = host_colormap;
	
	InitCommonControls();

	VID_SetPalette (palette);

	VID_SetMode ();

	GL_Init ();
}