/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "quakedef.h"
#include "winquake.h"
#include "menu.h"

#define	SLIDER_RANGE	10

/*
===============
M_DrawCharacter
===============
*/
void M_DrawCharacter (int cx, int line, int num)
{
	Draw_Character ( cx + ((vid.width - 400)>>1), line, num);
}

void M_Print (int cx, int cy, char *str)
{
	while (*str)
	{
		M_DrawCharacter (cx, cy, *str);
		str++;
		cx += 12;
	}
}

void M_DrawSlider (int x, int y, float range)
{
	int	i;

	if (range < 0)
		range = 0;
	if (range > 1)
		range = 1;
	M_DrawCharacter (x-8, y, 0);
	for (i=0 ; i<SLIDER_RANGE ; i++)
		M_DrawCharacter (x + i*8, y, 1);
	M_DrawCharacter (x+i*8, y, 2);
	M_DrawCharacter (x + (SLIDER_RANGE-1)*8 * range, y, 3);
}

void M_DrawCheckbox (int x, int y, int on)
{
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (on)
	{
		glColor3f (0,1,0);
		M_Print (x-8, y, "on");
	}
	else
	{
		glColor3f (1,0,0);
		M_Print (x-8, y, "off");
	}

	glColor3f(1,1,1);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

/*
================
M_ToggleMenu_f
================
*/

void M_ToggleMenu_f (void)
{
	m_entersound = true;

	if (key_dest == key_menu)
	{
		if (m_state != m_main)
		{
			M_Menu_Main_f ();
			return;
		}
		key_dest = key_game;
		m_state = m_none;
		return;
	}
	if (key_dest == key_console)
	{
		Con_ToggleConsole_f ();
	}
	else
	{
		M_Menu_Main_f ();
	}
}

void M_Init (void)
{
	Cmd_AddCommand ("togglemenu",		M_ToggleMenu_f);

	Cmd_AddCommand ("menu_main",		M_Menu_Main_f);
	Cmd_AddCommand ("menu_singleplayer",M_Menu_SinglePlayer_f);
	Cmd_AddCommand ("menu_load",		M_Menu_Load_f);
	Cmd_AddCommand ("menu_save",		M_Menu_Save_f);
	Cmd_AddCommand ("menu_multiplayer",	M_Menu_MultiPlayer_f);
	Cmd_AddCommand ("menu_setup",		M_Menu_Setup_f);
	Cmd_AddCommand ("menu_options",		M_Menu_Options_f);
	Cmd_AddCommand ("menu_controls",	M_Menu_Controls_f);
	Cmd_AddCommand ("menu_keys",		M_Menu_Keys_f);
	Cmd_AddCommand ("menu_sound",		M_Menu_Sound_f);
	Cmd_AddCommand ("menu_video",		M_Menu_Video_f);
	Cmd_AddCommand ("menu_misc",		M_Menu_Misc_f);
}

void M_Draw (void)
{
	if (m_state == m_none || key_dest != key_menu)
		return;

	if (!m_recursiveDraw)
	{

		if (scr_con_current)
		{
			Draw_ConsoleBackground (vid.height);
			S_ExtraUpdate ();
		}
		else
			Draw_FadeScreen ();
	}
	else
	{
		m_recursiveDraw = false;
	}

	switch (m_state)
	{
	case m_none:
		break;

	case m_main:
		M_Main_Draw ();
		break;

	case m_singleplayer:
		M_SinglePlayer_Draw ();
		break;

	case m_load:
		M_Load_Draw ();
		break;

	case m_save:
		M_Save_Draw ();
		break;

	case m_multiplayer:
		M_MultiPlayer_Draw ();
		break;

	case m_setup:
		M_Setup_Draw ();
		break;

	case m_net:
		M_Net_Draw ();
		break;

	case m_options:
		M_Options_Draw ();
		break;

	case m_controls:
		M_Controls_Draw ();
		break;

	case m_keys:
		M_Keys_Draw ();
		break;

	case m_sound:
		M_Sound_Draw ();
		break;

	case m_video:
		M_Video_Draw ();
		break;

	case m_misc:
		M_Misc_Draw ();
		break;

	case m_lanconfig:
		M_LanConfig_Draw ();
		break;

	case m_gameoptions:
		M_GameOptions_Draw ();
		break;

	case m_search:
		M_Search_Draw ();
		break;

	case m_slist:
		M_ServerList_Draw ();
		break;
	}

	if (m_entersound)
	{
		S_LocalSound ("misc/menu2.wav");
		m_entersound = false;
	}

	S_ExtraUpdate ();
}


void M_Keydown (int key)
{
	switch (m_state)
	{
	case m_none:
		return;

	case m_main:
		M_Main_Key (key);
		return;

	case m_singleplayer:
		M_SinglePlayer_Key (key);
		return;

	case m_load:
		M_Load_Key (key);
		return;

	case m_save:
		M_Save_Key (key);
		return;

	case m_multiplayer:
		M_MultiPlayer_Key (key);
		return;

	case m_setup:
		M_Setup_Key (key);
		return;

	case m_net:
		M_Net_Key (key);
		return;

	case m_options:
		M_Options_Key (key);
		return;

	case m_controls:
		M_Controls_Key (key);
		return;

	case m_keys:
		M_Keys_Key (key);
		return;

	case m_sound:
		M_Sound_Key (key);
		return;

	case m_video:
		M_Video_Key (key);
		return;

	case m_misc:
		M_Misc_Key (key);
		return;

	case m_lanconfig:
		M_LanConfig_Key (key);
		return;

	case m_gameoptions:
		M_GameOptions_Key (key);
		return;

	case m_slist:
		M_ServerList_Key (key);
		return;
	}
}