/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sbar.c -- status bar code

#include "quakedef.h"

qboolean	sb_showscores;

void Sbar_MiniDeathmatchOverlay (void);

/*
===============
Sbar_ShowScores
===============
*/
void Sbar_ShowScores (void)
{
	sb_showscores = true;
}

/*
===================
Sbar_DontShowScores
===================
*/
void Sbar_DontShowScores (void)
{
	sb_showscores = false;
}

/*
=========
Sbar_Init
=========
*/
void Sbar_Init (void)
{
	Cmd_AddCommand ("+showscores", Sbar_ShowScores);
	Cmd_AddCommand ("-showscores", Sbar_DontShowScores);
}

/*
==================
Sbar_DrawCharacter
==================
*/
void Sbar_DrawCharacter (int x, int y, int num)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_Character ( x + 4,							 y + vid.height-SBAR_HEIGHT, num);
	else
		Draw_Character ( x + 4 + ((vid.width - 320)>>1), y + vid.height-SBAR_HEIGHT, num);
}

//=============================================================================

int		fragsort[MAX_SCOREBOARD];

char	scoreboardtext[MAX_SCOREBOARD][20];
int		scoreboardtop[MAX_SCOREBOARD];
int		scoreboardbottom[MAX_SCOREBOARD];
int		scoreboardcount[MAX_SCOREBOARD];
int		scoreboardlines;

/*
==============
Sbar_SortFrags
==============
*/
void Sbar_SortFrags (void)
{
	int		i, j, k;

	scoreboardlines = 0;
	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.scores[i].name[0])
		{
			fragsort[scoreboardlines] = i;
			scoreboardlines++;
		}
	}

	for (i=0 ; i<scoreboardlines ; i++)
	{
		for (j=0 ; j<scoreboardlines-1-i ; j++)
		{
			if (cl.scores[fragsort[j]].frags < cl.scores[fragsort[j+1]].frags)
			{
				k = fragsort[j];
				fragsort[j]		= fragsort[j+1];
				fragsort[j+1]	= k;
			}
		}
	}
}

/*
=====================
Sbar_UpdateScoreboard
=====================
*/
void Sbar_UpdateScoreboard (void)
{
	int		i, k;
	int		top, bottom;
	scoreboard_t	*s;

	Sbar_SortFrags ();

	memset (scoreboardtext, 0, sizeof(scoreboardtext));

	for (i=0 ; i<scoreboardlines; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		sprintf (&scoreboardtext[i][1], "%3i %s", s->frags, s->name);

		top = s->colors & 0xf0;
		bottom = (s->colors & 15) <<4;
		scoreboardtop[i]	= 0;
		scoreboardbottom[i]	= 0;
	}
}

/*
===================
Sbar_DrawScoreboard
===================
*/
void Sbar_DrawScoreboard (void)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Sbar_DeathmatchOverlay ();
}

/*
=========
Sbar_Draw
=========
*/
void Sbar_Draw (void)
{
	if (scr_con_current == vid.height || cls.demoplayback)
		return;

	if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
	{
		Sbar_DrawScoreboard ();
	}
	else
	{
		Draw_String	(0,					vid.height - 48,	va("Ammo   %3i",	cl.stats[STAT_AMMO]),		(cl.stats[STAT_AMMO]	< 10));
		Draw_String (0,					vid.height - 32,	va("Hull   %3i",	cl.stats[STAT_HEALTH]),		(cl.stats[STAT_HEALTH]	< 25));
		Draw_String (0,					vid.height - 16,	va("Shield %3i",	cl.stats[STAT_ARMOR]),		(cl.stats[STAT_ARMOR]	< 25));
		Draw_String (vid.width - 96,	vid.height - 64,	va("Shield %i",		cl.stats[STAT_SHELLS]),		(cl.items & 3));
		Draw_String (vid.width - 96,	vid.height - 48,	va("Weapon %i",		cl.stats[STAT_NAILS]),		(cl.items & 12));
		Draw_String (vid.width - 96,	vid.height - 32,	va("Engine %i",		cl.stats[STAT_ROCKETS]),	(cl.items & 48));
		Draw_String (vid.width - 96,	vid.height - 16,	va("Repair %i",		cl.stats[STAT_CELLS]),		(cl.items & 64));

		if (cl.gametype == GAME_DEATHMATCH)
			Sbar_MiniDeathmatchOverlay ();
	}
}

//=============================================================================

/*
======================
Sbar_DeathmatchOverlay
======================
*/
void Sbar_DeathmatchOverlay (void)
{
	int				i, k, l;
	int				x, y, f;
	scoreboard_t	*s;

	Sbar_SortFrags ();

	l = scoreboardlines;

	x = 80 + ((vid.width - 320)>>1);
	y = 40;
	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

		f = s->frags;

		Draw_String (x + 12, y, va("%3i", f), false);

		if (k == cl.viewentity - 1)
			Draw_Character ( x - 12, y, 13);

		Draw_String (x + 64, y, s->name, false);

		y += 12;
	}
}

extern int notify_amount;

/*
==========================
Sbar_MiniDeathmatchOverlay
==========================
*/
void Sbar_MiniDeathmatchOverlay (void)
{
	int				i, k, l;
	int				y, f;
	scoreboard_t	*s;

	Sbar_SortFrags ();

	l = scoreboardlines;
	y = notify_amount;

	for (i = 0; i < scoreboardlines; i++)
		if (fragsort[i] == cl.viewentity - 1)
			break;

	if (i == scoreboardlines)
		i = 0;
	else
		i = i - 3;

	if (i > scoreboardlines - 3)
		i = scoreboardlines - 3;
	if (i < 0)
		i = 0;

	for (; i < scoreboardlines && y < (signed)vid.height - 8 ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;
		f = s->frags;

		if (k == cl.viewentity - 1) 
			Draw_String (0, y, va("[%i]", f), false);
		else
			Draw_String (12, y, va("%i", f), false);

		Draw_String (64, y, s->name, false);

		y += 16;
	}
}