/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// view.c -- player eye positioning

#include "quakedef.h"

cvar_t	crosshair = {"crosshair", "0", true};
cvar_t	cl_crossx = {"cl_crossx", "0", false};
cvar_t	cl_crossy = {"cl_crossy", "0", false};

//=============================================================================

cvar_t	v_centermove = {"v_centermove", "0.15", false};
cvar_t	v_centerspeed = {"v_centerspeed","500"};

void V_StartPitchDrift (void)
{
	if (cl.laststop == cl.time)
	{
		return;		// something else is keeping it from drifting
	}
	if (cl.nodrift || !cl.pitchvel)
	{
		cl.pitchvel = v_centerspeed.value;
		cl.nodrift = false;
		cl.driftmove = 0;
	}
}

void V_StopPitchDrift (void)
{
	cl.laststop = cl.time;
	cl.nodrift = true;
	cl.pitchvel = 0;
}

/*
===============
V_DriftPitch

Moves the client pitch angle towards cl.idealpitch sent by the server.

If the user is adjusting pitch manually, either with lookup/lookdown,
mlook and mouse, or klook and keyboard, pitch drifting is constantly stopped.

Drifting is enabled when the center view key is hit, mlook is released and
lookspring is non 0, or when 
===============
*/
void V_DriftPitch (void)
{
	float		delta, move;

	if (noclip_anglehack || !cl.onground || cls.demoplayback )
	{
		cl.driftmove = 0;
		cl.pitchvel = 0;
		return;
	}

// don't count small mouse motion
	if (cl.nodrift)
	{
		if ( fabs(cl.cmd.forwardmove) < cl_forwardspeed.value)
			cl.driftmove = 0;
		else
			cl.driftmove += host_frametime;
	
		if ( cl.driftmove > v_centermove.value)
		{
			V_StartPitchDrift ();
		}
		return;
	}
	
	delta = cl.idealpitch - cl.viewangles[PITCH];

	if (!delta)
	{
		cl.pitchvel = 0;
		return;
	}

	move = host_frametime * cl.pitchvel;
	cl.pitchvel += host_frametime * v_centerspeed.value;
	
//Con_Printf ("move: %f (%f)\n", move, host_frametime);

	if (delta > 0)
	{
		if (move > delta)
		{
			cl.pitchvel = 0;
			move = delta;
		}
		cl.viewangles[PITCH] += move;
	}
	else if (delta < 0)
	{
		if (move > -delta)
		{
			cl.pitchvel = 0;
			move = -delta;
		}
		cl.viewangles[PITCH] -= move;
	}
}

/*
============================================================================== 
 
						PALETTE FLASHES 
 
============================================================================== 
*/ 

byte		gammatable[256];	// palette is sent through this

byte		ramps[3][256];
float		v_blend[4];		// rgba 0.0 - 1.0

void BuildGammaTable (float g)
{
	int		i, inf;
	
	if (g == 1.0)
	{
		for (i=0 ; i<256 ; i++)
			gammatable[i] = i;
		return;
	}
	
	for (i=0 ; i<256 ; i++)
	{
		inf = 255 * pow ( (i+0.5)/255.5 , g ) + 0.5;
		if (inf < 0)
			inf = 0;
		if (inf > 255)
			inf = 255;
		gammatable[i] = inf;
	}
}

/*
===============
V_ParseDamage
===============
*/
void V_ParseDamage (void)
{
	int		armor, blood;
	vec3_t	from;
	float	count;
	
	armor = MSG_ReadByte ();
	blood = MSG_ReadByte ();

	from[0] = MSG_ReadCoord ();
	from[1] = MSG_ReadCoord ();
	from[2] = MSG_ReadCoord ();

	if (!armor && !blood)
	{
		cl.cshift.percent		= 0;
		cl.cshift.destcolor[0]	= 0;
		cl.cshift.destcolor[1]	= 0;
		cl.cshift.destcolor[2]	= 0;
		return;
	}

	count = blood*0.5 + armor*0.5;
	if (count < 10)
		count = 10;

	cl.faceanimtime = cl.time + 0.2;		// but sbar face into pain frame

	cl.cshift.percent += 3*count;
	if (cl.cshift.percent < 0)
		cl.cshift.percent = 0;
	if (cl.cshift.percent > 150)
		cl.cshift.percent = 150;

	if (armor > blood)		
	{
		cl.cshift.destcolor[0] = 200;
		cl.cshift.destcolor[1] = 100;
		cl.cshift.destcolor[2] = 100;
	}
	else if (armor)
	{
		cl.cshift.destcolor[0] = 220;
		cl.cshift.destcolor[1] = 50;
		cl.cshift.destcolor[2] = 50;
	}
	else
	{
		cl.cshift.destcolor[0] = 255;
		cl.cshift.destcolor[1] = 0;
		cl.cshift.destcolor[2] = 0;
	}
}


/*
==================
V_cshift_f
==================
*/
void V_cshift_f (void)
{
}


/*
==============
V_BonusFlash_f
==============
*/
void V_BonusFlash_f (void)
{
}

/*
=============
V_CalcBlend
=============
*/
void V_CalcBlend (void)
{
	float	r, g, b, a, a2;

	r = 0;
	g = 0;
	b = 0;
	a = 0;

	a2 = cl.cshift.percent * 0.0039215686;	// Tomaz - Speed

	if (!a2)
	{
		v_blend[3] = 0;
		return;
	}

	a	= a + a2 * (1-a);
	a2	= a2/a;

	r	= (1-a2) + cl.cshift.destcolor[0] * a2;
	g	= (1-a2) + cl.cshift.destcolor[1] * a2;
	b	= (1-a2) + cl.cshift.destcolor[2] * a2;

	v_blend[0] = r * 0.0039215686;	// Tomaz - Speed
	v_blend[1] = g * 0.0039215686;	// Tomaz - Speed
	v_blend[2] = b * 0.0039215686;	// Tomaz - Speed
	v_blend[3] = a;
	if (v_blend[3] > 1)
		v_blend[3] = 1;
	if (v_blend[3] < 0)
		v_blend[3] = 0;
}

/*
===============
V_UpdatePalette
===============
*/
void V_UpdatePalette (void)
{
	int		i;
	byte	*basepal, *newpal;
	byte	pal[768];
	float	r,g,b,a;
	int		ir, ig, ib;

// drop the damage value
	cl.cshift.percent -= host_frametime*150;
	if (cl.cshift.percent <= 0)
		cl.cshift.percent = 0;

	V_CalcBlend ();

	a = v_blend[3];
	r = 255*v_blend[0]*a;
	g = 255*v_blend[1]*a;
	b = 255*v_blend[2]*a;

	a = 1-a;
	for (i=0 ; i<256 ; i++)
	{
		ir = i*a + r;
		ig = i*a + g;
		ib = i*a + b;
		if (ir > 255)
			ir = 255;
		if (ig > 255)
			ig = 255;
		if (ib > 255)
			ib = 255;

		ramps[0][i] = gammatable[ir];
		ramps[1][i] = gammatable[ig];
		ramps[2][i] = gammatable[ib];
	}

	basepal = host_basepal;
	newpal = pal;
	
	for (i=0 ; i<256 ; i++)
	{
		ir = basepal[0];
		ig = basepal[1];
		ib = basepal[2];
		basepal += 3;
		
		newpal[0] = ramps[0][ir];
		newpal[1] = ramps[1][ig];
		newpal[2] = ramps[2][ib];
		newpal += 3;
	}
}

/* 
============================================================================== 
 
						VIEW RENDERING 
 
============================================================================== 
*/ 

/*
==================
V_CalcIntermissionRefdef
==================
*/
void V_CalcIntermissionRefdef (void)
{
	entity_t	*ent, *view;

	ent		= &cl_entities[cl.viewentity];
	view	= &cl.viewent;

	r_refdef.angles[0]	= ent->angles[0];
	r_refdef.angles[1]	= ent->angles[1];
	r_refdef.angles[2]	= ent->angles[2];

	r_refdef.origin[0]	= ent->origin[0];
	r_refdef.origin[1]	= ent->origin[1];
	r_refdef.origin[2]	= ent->origin[2];

	view->model = NULL;
}

/*
============
V_CalcRefdef
============
*/
void V_CalcRefdef (void)
{
	entity_t	*ent, *view;

	ent		= &cl_entities[cl.viewentity];
	view	= &cl.viewent;
	
	ent->angles[0]		= -cl.viewangles[0];
	ent->angles[1]		=  cl.viewangles[1];
	ent->angles[2]		=  cl.viewangles[2];

	r_refdef.angles[0]	= cl.viewangles[0];
	r_refdef.angles[1]	= cl.viewangles[1];
	r_refdef.angles[2]	= cl.viewangles[2];

	r_refdef.origin[0]	= ent->origin[0];
	r_refdef.origin[1]	= ent->origin[1];
	r_refdef.origin[2]	= ent->origin[2] + cl.viewheight;
			
// set up gun position

	view->angles[0]		= ent->angles[0];
	view->angles[1]		= ent->angles[1];
	view->angles[2]		= ent->angles[2];

	view->origin[0]		= ent->origin[0];
	view->origin[1]		= ent->origin[1];
	view->origin[2]		= ent->origin[2];

	view->model			= cl.model_precache[cl.stats[STAT_WEAPON]];
	view->frame			= cl.stats[STAT_WEAPONFRAME];
	view->colormap		= vid.colormap;
}

/*
============
V_RenderView
============
*/

extern qboolean changed_res;

void V_RenderView (void)
{
	if (con_forcedup)
		return;

	if (changed_res)
		changed_res = false;

	if (cl.intermission)
	{	// intermission / finale rendering
		V_CalcIntermissionRefdef ();	
	}
	else
	{
		if (!cl.paused /* && (sv.maxclients > 1 || key_dest == key_game) */ )
			V_CalcRefdef ();
	}

	R_PushDlights ();

	R_RenderView ();
}

//============================================================================

/*
=============
V_Init
=============
*/
void V_Init (void)
{
	Cmd_AddCommand ("v_cshift",		V_cshift_f);	
	Cmd_AddCommand ("bf",			V_BonusFlash_f);
	Cmd_AddCommand ("centerview",	V_StartPitchDrift);

	Cvar_RegisterVariable (&v_centermove);
	Cvar_RegisterVariable (&v_centerspeed);

	Cvar_RegisterVariable (&crosshair);
	Cvar_RegisterVariable (&cl_crossx);
	Cvar_RegisterVariable (&cl_crossy);
	
	BuildGammaTable (1.0);	// no gamma yet
}